import logging.config
import os
from logging.handlers import WatchedFileHandler

LOG_FILE = '/var/log/cloudlinux-backup-utils.log'

DEBUG = 3
VERBOSE = 2
INFO = 1
CONCISE = 0

_format_concise = logging.Formatter('%(message)s')
_format_verbose = logging.Formatter(
    '%(levelname)-7s [%(asctime)s] %(name)s: %(message)s'
)

_config = {
    CONCISE: [_format_concise, logging.WARNING],
    INFO: [_format_concise, logging.INFO],
    VERBOSE: [_format_verbose, logging.INFO],
    DEBUG: [_format_verbose, logging.DEBUG],
}

MAIN_LOGGER = 'restore_infected'
CLI_LOGGER = 'restore_infected.cli'

FILE_HANDLER = 'file'
CONSOLE_HANDLER = 'console'
HANDLERS = {}

ALL_LOGGERS = [
    MAIN_LOGGER,
    CLI_LOGGER,
    'AcronisClientInstaller',
    'requests',
]


def log_only_cli_module():
    disable_handler(FILE_HANDLER)
    ALL_LOGGERS.clear()

    logger = logging.getLogger(CLI_LOGGER)
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    handler.setFormatter(_format_concise)
    logger.addHandler(handler)


def log_to_file(output=None):
    disable_handler(FILE_HANDLER)
    if output:
        init_file_handler(output)
        enable_handler(FILE_HANDLER)


def log_to_console(verbosity):
    if verbosity > CONCISE:
        init_console_handler(verbosity)
        enable_handler(CONSOLE_HANDLER)
    else:
        disable_handler(CONSOLE_HANDLER)


def set_config(handler, verbosity):
    fmt, lvl = _config.get(verbosity, _config[CONCISE])
    handler.setFormatter(fmt)
    handler.setLevel(lvl)


def init_file_handler(path):
    if os.access(os.path.dirname(path), os.W_OK | os.X_OK):
        handler = WatchedFileHandler(path, encoding='utf8')
    else:
        handler = logging.NullHandler()

    set_config(handler, DEBUG)
    HANDLERS[FILE_HANDLER] = handler
    return handler


def init_console_handler(verbosity):
    handler = logging.StreamHandler()
    set_config(handler, verbosity)
    HANDLERS[CONSOLE_HANDLER] = handler
    return handler


def enable_handler(handler_name):
    handler = HANDLERS.get(handler_name)
    if not handler and handler_name == FILE_HANDLER:
        handler = init_file_handler(LOG_FILE)
    elif not handler:
        handler = init_console_handler(verbosity=CONCISE)

    for logger_name in ALL_LOGGERS:
        logger = logging.getLogger(logger_name)
        logger.addHandler(handler)
    HANDLERS[handler_name] = handler


def disable_handler(handler_name):
    handler = HANDLERS.get(handler_name)
    for logger_name in ALL_LOGGERS:
        logger = logging.getLogger(logger_name)
        logger.addHandler(handler)
    HANDLERS.pop(handler_name, None)
